// -*- C++ -*-
/*!
 * @file  SmartPal_controller.cpp
 * @brief SmartPal controller
 * @date $Date$
 *
 * $Id$
 */

#include "SmartPal_controller.h"
#include <math.h>

using namespace boost::numeric::ublas;

#define DOF 23
#define input_num 23
#define WAIST_ID 0
#define WAIST_num 2
#define RARM_ID 2
#define RARM_num 7
#define LARM_ID 10
#define LARM_num 7
#define WHEEL_ID 18
#define WHEEL_num 4
#define HAND_ID 9
#define HAND_num 1
#define HEAD_ID 22
#define HEAD_num 2
#define ENDEFFECTOR_ID 9
#define r 0.1 // center to wheel
#define L 0.2495 // wheel radius
#define Cx 0
#define Cy 0
#define TIMESTEP 0.001
#define Init_Pose_Time 0.05
#define GAIN_FILE  "PDgain.dat"
#define MODEL_URL  "file:../../model/SmartPal5/SmartPal5/SmartPal5_withODV_withCamera.wrl"
#define NS_OPT "-ORBInitRef NameService=corbaloc:iiop:localhost:2809/NameService"
#define Image_Size 640*480

double init_pose_count = 0;

#define DEG2RAD(x)  ((x)/180.0*3.1415)
#define RAD2DEG(x)  ((x)/3.1415*180.0)


//#define LOGGING
#ifdef LOGGING
std::ofstream ofs1("log_input.dat");
std::ofstream ofs2("log_q_ref.dat");
std::ofstream ofs3("log_output.dat");
std::ofstream ofs4("log_q_cur.dat");
std::ofstream ofs5("log_q_ref_out.dat");
#endif //LOGGING



// Module specification
// <rtc-template block="module_spec">
static const char* Pal_controller_spec[] =
  {
    "implementation_id", "SmartPal_controller",
    "type_name",         "SmartPal_controller",
    "description",       "SmartPal controller",
    "version",           "1.0.0",
    "vendor",            "nomura",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
SmartPal_controller::SmartPal_controller(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_q_cur_inIn("q_cur_in", m_q_cur_in),
    m_q_arm_ref_inIn("q_arm_ref_in", m_q_arm_ref_in),
    m_q_ref_outOut("q_ref_out", m_q_ref_out),
    m_dq_ref_outOut("dq_ref_out", m_ddq_ref_out),
    m_ddq_ref_outOut("ddq_ref_out", m_ddq_ref_out),
    m_q_cur_outOut("q_cur_out", m_q_cur_out),

    q_cur(DOF),
    q_ref(DOF),
    dq_cur(DOF),
    dq_ref(DOF),
    q_cur_old(DOF),
    q_ref_old(DOF),
    q_init(DOF)

    // </rtc-template>
{

  Pgain = new double[DOF];
  Dgain = new double[DOF];
  
  if (access(GAIN_FILE, 0)){
    std::cerr << GAIN_FILE << " not found" << std::endl;
  }else{
    gain.open(GAIN_FILE);
    for (int i=0; i<DOF; i++){
      gain >> Pgain[i];
      gain >> Dgain[i];
    }
    gain.close();
  }

}

/*!
 * @brief destructor
 */
SmartPal_controller::~SmartPal_controller()
{
  delete [] Pgain;
  delete [] Dgain;
}



RTC::ReturnCode_t SmartPal_controller::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("q_cur_in", m_q_cur_inIn);
  addInPort("q_arm_ref_in", m_q_arm_ref_inIn);
  
  // Set OutPort buffer
  addOutPort("q_ref_out", m_q_ref_outOut);
  addOutPort("dq_ref_out", m_dq_ref_outOut);
  addOutPort("ddq_ref_out", m_ddq_ref_outOut);
  addOutPort("q_cur_out", m_q_cur_outOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Pgain", m_Pgain, "40000");
  bindParameter("Dgain", m_Dgain, "100");

  bindParameter("str_model_file", m_str_model_file, MODEL_URL);


  // </rtc-template>

  m_q_arm_ref_in.data.length(input_num);
  m_q_cur_in.data.length(DOF);
  m_q_ref_out.data.length(DOF);
  m_dq_ref_out.data.length(DOF);
  m_ddq_ref_out.data.length(DOF);
  m_q_cur_out.data.length(input_num);

  std::cout << "---------- on Initialize ----------" << std::endl;

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SmartPal_controller::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t SmartPal_controller::onActivated(RTC::UniqueId ec_id)
{

  m_q_arm_ref_inIn.update();
  m_q_cur_inIn.update();
  
  for(int i=0; i < DOF; ++i){
    q_cur_old[i] = m_q_cur_in.data[i];
    q_ref_old[i] = 0;
    q_cur[i] = m_q_cur_in.data[i];
    q_ref[i] = 0;
    m_dq_ref_out.data[i] = 0;
    m_ddq_ref_out.data[i] = 0;
  }
  

  // get current directory
  char current_dir[MAXPATHLEN];
  char *p = NULL;
  
  p = getcwd (current_dir, MAXPATHLEN);
  if(p == NULL) {
    perror("getcwd");
    return RTC::RTC_ERROR;
  }
  
  
  // model load //
  std::stringstream url;
  url << "file://";
  url << current_dir;
  url << "/";
  url << m_str_model_file;

  std::istringstream OPT(NS_OPT);
  body = loadBodyFromModelLoader(url.str().c_str(),OPT);
  if(body==NULL) {
    std::cerr << "ModelLoader couldn't load " << url.str().c_str() << std::endl;
  }
  
  std::cout << "---------- on Activated ----------" << std::endl;

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SmartPal_controller::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t SmartPal_controller::onExecute(RTC::UniqueId ec_id)
{

  m_q_arm_ref_inIn.update();
  m_q_cur_inIn.update();

  /////////////// q_cur limit ///////////////

  //  for(int i=0; i < RARM_num; ++i){
  for(int i=0; i < WHEEL_ID; ++i){
    if(m_q_cur_in.data[RARM_ID+i]>3.14 || m_q_cur_in.data[RARM_ID+i]<-3.14){
      return RTC::RTC_OK;
    }
  }

  /////////////// Translate input[deg] -> q_ref[rad] ///////////////

  ///// Translate WAIST /////
  q_ref[WAIST_ID]   = DEG2RAD(m_q_arm_ref_in.data[4]) * -1.0 / 3.0;
  q_ref[WAIST_ID+1] = DEG2RAD(m_q_arm_ref_in.data[4]) *  4.0 / 3.0;

  ///// Translate RARM /////
  for(int i=0; i < RARM_num; ++i){
    q_ref[RARM_ID+i] = DEG2RAD(m_q_arm_ref_in.data[6+i]);//input[RARM]=6
  }

  ///// Translate LARM /////
  for(int i=0; i < LARM_num; ++i){
    q_ref[LARM_ID+i] = DEG2RAD(m_q_arm_ref_in.data[13+i]);//input[LARM]=13
  }

  ///// Translate HEAD /////
  for(int i=0; i < HEAD_num; ++i){
    q_ref[HEAD_ID+i] = DEG2RAD(m_q_arm_ref_in.data[20+i]);//input[HEAD]=20
  }

  /////////////// Controll SmartPal5 ///////////////

  try {

    ////////// ARM Angle Initialize//////////
    if(init_pose_count < Init_Pose_Time){
      
      if(init_pose_count == 0){
	for(int i=0; i < DOF; ++i){
	  q_init[i] = m_q_cur_in.data[i];
	}
      }

      for(int i=0; i < RARM_num; ++i){
      	q_ref[RARM_ID+i] = q_init[RARM_ID+i];
      }

      init_pose_count = init_pose_count + TIMESTEP;
    }

    ////////// ARM ANGLE LIMITER //////////
    for(int i=0; i < RARM_num; ++i){
      if(q_ref[RARM_ID+i]-m_q_cur_in.data[RARM_ID+i]>0.3 || 
	 q_ref[RARM_ID+i]-m_q_cur_in.data[RARM_ID+i]<-0.3)
	{
	  //std::cerr << "[Controller] velocity limit over  " << i << "  "  << q_ref[RARM_ID+i]-m_q_cur_in.data[RARM_ID+i] << std::endl;
	  for(int j=0; j < RARM_num; ++j){
	    q_ref[RARM_ID+j] = q_ref_old[RARM_ID+j];
	  }
	}
    }
    
    ////////// Controlle HAND //////////
    for(int i=0; i < HAND_num; ++i){
      q_ref[HAND_ID + i] = q_ref_old[HAND_ID + i];
    }

    for(int i=0; i<DOF; i++){
      q_cur[i] = m_q_cur_in.data[i];
      dq_ref[i] = (q_ref[i] - q_ref_old[i]) / TIMESTEP;
      dq_cur[i] = (q_cur[i] - q_cur_old[i]) / TIMESTEP;
      q_ref_old[i] = q_ref[i];
      q_cur_old[i] = q_cur[i];
      m_q_ref_out.data[i] = q_ref[i];
      m_dq_ref_out.data[i] = 0;
      m_ddq_ref_out.data[i] = 0;
    }
    
    ////////// Controlle ODV //////////
        
    V_x     = m_q_arm_ref_in.data[0] - x_old;
    V_y     = m_q_arm_ref_in.data[1] - y_old;
    V_theta = m_q_arm_ref_in.data[2] - theta_old;
    
    x_old     = m_q_arm_ref_in.data[0];
    y_old     = m_q_arm_ref_in.data[1];
    theta_old = m_q_arm_ref_in.data[2];
    
    dq_ref[WHEEL_ID+0] = -( V_y + (L - Cy) * V_theta) / r;
    dq_ref[WHEEL_ID+1] = -( V_x + (L - Cx) * V_theta) / r;
    dq_ref[WHEEL_ID+2] = -(-V_y + (L - Cy) * V_theta) / r;
    dq_ref[WHEEL_ID+3] = -(-V_x + (L - Cx) * V_theta) / r;
    
    for(int i=0; i < WHEEL_num; i++){
      m_q_ref_out.data[WHEEL_ID + i] = dq_ref[WHEEL_ID + i] * Pgain[WHEEL_ID + i];
    }

  }

  catch(...) {
    
    std::cerr << "an error occured" << std::endl;
        
    for(int i=0; i<DOF; i++){
      q_ref[i] = q_ref_old[i];
      q_cur[i] = m_q_cur_in.data[i];
      dq_ref[i] = (q_ref[i] - q_ref_old[i]) / TIMESTEP;
      dq_cur[i] = (q_cur[i] - q_cur_old[i]) / TIMESTEP;
      q_ref_old[i] = q_ref[i];
      q_cur_old[i] = q_cur[i];
      m_q_ref_out.data[i] = q_ref[i];
      m_dq_ref_out.data[i] = 0;
      m_ddq_ref_out.data[i] = 0;
    }

    for(int i=0; i < WHEEL_num; i++){
      m_q_ref_out.data[WHEEL_ID + i] = q_cur[WHEEL_ID + i];
    }
  }

  /////////////// Q_REF_OUT LIMIT ///////////////

  for(int j=0; j<DOF; j++){
    if(m_q_ref_out.data[j] > 3.14 || m_q_ref_out.data[j] < -3.14){
      for(int i=0; i<DOF; i++){
	m_q_ref_out.data[i] = m_q_cur_in.data[i];
      }
    }
  }

  /////////////// Translate q_cur[rad] -> output[deg] ///////////////

  ///// Translate WAIST /////
  for(int i=0; i < WAIST_num; ++i){
    m_q_cur_out.data[3+i] = 0.0;//output[WAIST]=3
  }
  m_q_cur_out.data[4] = RAD2DEG(q_cur[WAIST_ID] + q_cur[WAIST_ID+1]);

  ///// Translate RARM /////
  for(int i=0; i < RARM_num; ++i){
    m_q_cur_out.data[6+i] = RAD2DEG(q_cur[RARM_ID+i]);//output[RARM]=6
  }

  ///// Translate LARM /////
  for(int i=0; i < LARM_num; ++i){
    m_q_cur_out.data[13+i] = RAD2DEG(q_cur[LARM_ID+i]);//output[LARM]=13
  }

  ///// Translate HEAD /////
  for(int i=0; i < HEAD_num; ++i){
    m_q_cur_out.data[20+i] = RAD2DEG(q_cur[HEAD_ID+i]);//output[HEAD]=20
  }

  ///// Translate ODV /////
  for(int i=0; i < 3; ++i){
    m_q_cur_out.data[i] = 0;
  }

  /////////////// OUTPUT ///////////////
  
  m_q_ref_outOut.write();
  m_dq_ref_outOut.write();
  m_ddq_ref_outOut.write();
  m_q_cur_outOut.write();
 

#ifdef LOGGING
  /////////////// LOG ///////////////
  ofs1.precision(3);
  ofs1 << m_q_arm_ref_in.data[0];
  for(int i=1; i<input_num; i++){
    ofs1 << ", " << m_q_arm_ref_in.data[i];
  }
  ofs1 << std::endl;

  ofs2.precision(3);
  ofs2 << q_ref[0];
  for(int i=1; i<DOF; i++){
    ofs2 << ", " << q_ref[i];
  }
  ofs2 << std::endl;

  ofs3.precision(3);
  ofs3 << m_q_cur_out.data[0];
  for(int i=1; i<input_num; i++){
    ofs3 << ", " << m_q_cur_out.data[i];
  }
  ofs3 << std::endl;

  ofs4.precision(3);
  ofs4 << q_cur[0];
  for(int i=1; i<DOF; i++){
    ofs4 << ", " << q_cur[i];
  }
  ofs4 << std::endl;

  ofs5.precision(3);
  ofs5 << m_q_ref_out.data[0];
  for(int i=1; i<DOF; i++){
    ofs5 << ", " << m_q_ref_out.data[i];
  }
  ofs5 << std::endl;
#endif

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SmartPal_controller::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SmartPal_controller::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void SmartPal_controllerInit(RTC::Manager* manager)
  {
    coil::Properties profile(Pal_controller_spec);
    manager->registerFactory(profile,
                             RTC::Create<SmartPal_controller>,
                             RTC::Delete<SmartPal_controller>);
  }
  
};


