// -*- C++ -*-
/*!
 * @file  RedundancySolutionCore.cpp
 * @brief ModuleDescription
 * @date $Date$
 *
 * $Id$
 */

#include "RedundancySolutionCore.h"


//#define LOGGING
#ifdef LOGGING
std::ofstream ofs1("log_xyzR_in.dat");
std::ofstream ofs2("log_q_ref_out.dat");
std::ofstream ofs3("log_q_cur_in.dat");
std::ofstream ofs4("log_xyzR_out.dat");
std::ofstream ofs5("log_redun_ref_in.dat");
std::ofstream ofs6("log_redun_cur_out.dat");
#endif //LOGGING



// Module specification
static const char* redundancysolutioncore_spec[] =
  {
    "implementation_id", "RedundancySolutionCore",
    "type_name",         "RedundancySolutionCore",
    "description",       "ModuleDescription",
    "version",           "1.0.0",
    "vendor",            "tohoku",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
    "conf.default.conf_name"             ,  "robot_conf",
    "conf.default.double_dt"             ,  "0.001",
    "conf.default.double_limit_per_sec"  ,  "6.28318",
    // Widget
    "conf.__widget__.conf_name",            "text",
    "conf.__widget__.double_dt",            "text",
    "conf.__widget__.double_limit_per_sec", "text",
    ""
  };



// constructor
RedundancySolutionCore::RedundancySolutionCore(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_q_cur_inIn("q_cur_in", m_q_cur_in),
    m_xyzR_inIn("xyzR_in", m_xyzR_in),
    m_redun_ref_inIn("redun_ref_in", m_redun_ref_in),
    m_q_ref_outOut("q_ref_out", m_q_ref_out),
    m_xyzR_outOut("xyzR_out", m_xyzR_out),
    m_redun_cur_outOut("redun_cur_out", m_redun_cur_out),
    m_status_outOut("status_out", m_status_out),
    m_toRedundancyPort("toRedundancy")
{
  CALLBACK_CLASS_INIT(m_q_cur_in);
  CALLBACK_CLASS_INIT(m_xyzR_in);

  m_toRedundancy.init(this);
  firstCircle = true;
}


// destructor
RedundancySolutionCore::~RedundancySolutionCore()
{
  CALLBACK_CLASS_CLEAR(m_q_cur_in);
  CALLBACK_CLASS_CLEAR(m_xyzR_in);
}



RTC::ReturnCode_t RedundancySolutionCore::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // Set InPort buffers
  addInPort("q_cur_in", m_q_cur_inIn);
  addInPort("xyzR_in", m_xyzR_inIn);
  addInPort("redun_ref_in", m_redun_ref_inIn);
  
  // Set OutPort buffer
  addOutPort("q_ref_out", m_q_ref_outOut);
  addOutPort("xyzR_out", m_xyzR_outOut);
  addOutPort("redun_cur_out", m_redun_cur_outOut);
  addOutPort("status_out", m_status_outOut);
  
  // Set service provider to Ports
  m_toRedundancyPort.registerProvider("toRedundancy", "RedundancyConfig", m_toRedundancy);

  // Set CORBA Service Ports
  addPort(m_toRedundancyPort);


  // Bind variables and configuration variable
  bindParameter("double_dt", m_double_dt, "0.001");
  bindParameter("double_limit_per_sec", m_double_limit, "6.28318");
  bindParameter("conf_name", m_conf_name, "robot_conf");


  //------------------------------------------------------------------------------------
  std::cout << "---------- Redundancy on Initialized ----------" << std::endl;
  return RTC::RTC_OK;
}



RTC::ReturnCode_t RedundancySolutionCore::onActivated(RTC::UniqueId ec_id)
{
  //------------------------------------------------------------------------------------
  //  call back
  //------------------------------------------------------------------------------------
  CALLBACK_CLASS_SET(m_q_cur_in);
  CALLBACK_CLASS_SET(m_xyzR_in);


  //------------------------------------------------------------------------------------
  //  data port
  //------------------------------------------------------------------------------------
  // 入出力ポートの設定
  m_xyzR_out.data.length(24);
  m_xyzR_in.data.length(24);

  m_q_ref_out.data.length(23);
  m_q_cur_in.data.length(23);

  m_redun_cur_out.data.length(8);
  m_redun_ref_in.data.length(8);

  m_status_out.data.length(2);


  // to keep reference data
  m_xyzR_in_keeper.data.length(24);
  m_redun_ref_in_keeper.data.length(8);

  
  // デフォルト値の設定
  for(unsigned int i = 0; i < 24; i++) {
    m_xyzR_out.data[i] = 0.0;
    m_xyzR_in.data[i]  = 0.0;

    m_xyzR_in_keeper.data[i] = 0.0;
  }
  for(unsigned int i = 0; i < 23; i++) {
    m_q_ref_out.data[i] = 0.0;
    m_q_cur_in.data[i]  = 0.0;
  }
  for(unsigned int i = 0; i < 8; i++) {
    m_redun_cur_out.data[i] = 0.0;
    m_redun_ref_in.data[i]  = 0.0;

    m_redun_ref_in_keeper.data[i] = 0.0;
  }
  m_status_out.data[0] = 0.0;  m_status_out.data[1] = 0.0;


  //------------------------------------------------------------------------------------
  //  configuration set
  //------------------------------------------------------------------------------------
  m_configsets.update();  
  RTC::Properties conf = m_configsets.getConfigurationSet(m_conf_name.c_str());
  jointname.setConf(conf);


  //------------------------------------------------------------------------------------
  //  engine
  //------------------------------------------------------------------------------------
  // init engine
  if(engine.init(&jointname) == false) {
    cout << "error : [Redun] : Engine init false" << endl;
    return RTC::RTC_ERROR;
  }

  // SC法のリミッター設定
  engine.setProperty(m_double_dt, m_double_limit);
  

  //engine.test();

  //------------------------------------------------------------------------------------
  std::cout << "---------- Redundancy on Activated ----------" << std::endl;
  return RTC::RTC_OK;
}




RTC::ReturnCode_t RedundancySolutionCore::onDeactivated(RTC::UniqueId ec_id)
{
  engine.finish();
  jointname.initParameter();

  std::cout << "---------- Redundancy on Deactivated ----------" << std::endl;
  return RTC::RTC_OK;
}




RTC::ReturnCode_t RedundancySolutionCore::onExecute(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


////////////////////////////////////////////////////////////////////////////////////////

// for service port
void RedundancySolutionCore::selectMode()
{
  // 目標値を現在値に設定
  m_xyzR_in_keeper = engine.getCurPosition();
  m_redun_ref_in_keeper = engine.getCurRedundant();
}



// 目標位置のセット
void RedundancySolutionCore::setRefPosition(const TimedDoubleSeq& d)
{
  if(m_toRedundancy.m_mode == 0 || m_toRedundancy.m_mode == 2) {

    //------------------------------------------------------------------------------------
    //  単位変換
    //------------------------------------------------------------------------------------    
    switch(m_toRedundancy.m_type) {
      
    case RIGHT: {
      for(unsigned int i = 0; i < 3; i++) {
	m_xyzR_in_keeper.data[i]    = d.data[i] / 1000.0;     // 右腕（位置） mm -> m
	 }
      for(unsigned int i = 3; i < 12; i++) {
	m_xyzR_in_keeper.data[i]    = d.data[i];              // 右腕（姿勢）
      }
      break;
    } // end of case RIGHT
      
      
    case LEFT: {
      for(unsigned int i = 0; i < 3; i++) {
	m_xyzR_in_keeper.data[i+12] = d.data[i+12] / 1000.0;  // 左腕（位置） mm -> m
      }
      for(unsigned int i = 3; i < 12; i++) {
	m_xyzR_in_keeper.data[i+12] = d.data[i+12];           // 左腕（姿勢）
      }
      break;
    } // end of case LEFT
      
      
    case DUAL: {
      for(unsigned int i = 0; i < 3; i++) {
	m_xyzR_in_keeper.data[i]    = d.data[i] / 1000.0;     // 右腕（位置） mm -> m
	m_xyzR_in_keeper.data[i+12] = d.data[i+12] / 1000.0;  // 左腕（位置） mm -> m
      }
      for(unsigned int i = 3; i < 12; i++) {
	m_xyzR_in_keeper.data[i]    = d.data[i];              // 右腕（姿勢）
	m_xyzR_in_keeper.data[i+12] = d.data[i+12];           // 左腕（姿勢）
      }
    } // end of case DUAL
      
    } // end of switch
    
  } // end of if



  //------------------------------------------------------------------------------------
  //  目標値
  //------------------------------------------------------------------------------------
  if(engine.setRefPosition(m_xyzR_in_keeper) == false) {
    cout << "error : [Redun] : setRefPosition" << endl;
    //return;
  }
}



// 冗長軸の目標値セット
void RedundancySolutionCore::setRefRedundant(const TimedDoubleSeq& phi)
{
  if(m_toRedundancy.m_mode == 1 || m_toRedundancy.m_mode == 2) {
    
    //------------------------------------------------------------------------------------
    //  単位変換
    //------------------------------------------------------------------------------------    
    for(int i = 0; i < 2; i++) {
      m_redun_ref_in_keeper.data[i] = phi.data[i] / 1000.0;
    }
    for(int i = 2; i < 8; i++) {
      m_redun_ref_in_keeper.data[i] = DEG2RAD(phi.data[i]);
    }

    // engineではWAIST部分がZYXの順番なので逆転
    // (rx, ry, rz) -> (rz, ry, rx)
    double rx, ry, rz;
    rx = m_redun_ref_in_keeper.data[3];    ry = m_redun_ref_in_keeper.data[4];    rz = m_redun_ref_in_keeper.data[5];
    m_redun_ref_in_keeper.data[3] = rz;    m_redun_ref_in_keeper.data[4] = ry;    m_redun_ref_in_keeper.data[5] = rx;
    
  } // end of if



  //------------------------------------------------------------------------------------
  //  目標値
  //------------------------------------------------------------------------------------
  if(engine.setRefRedundant(m_redun_ref_in_keeper) == false) {
    cout << "error : [Redun] : setRefRedundant" << endl;
    //return;
  }     
}



////////////////////////////////////////////////////////////////////////////////////////

void RedundancySolutionCore::_update_m_q_cur_in(const ConnectorInfo& info, const TimedDoubleSeq& q)
{
  if(q.data.length() != 23) {
    cout << "error : [Redun] : update m_q_cur_in" << endl;
    return;
  }
  
  
  //------------------------------------------------------------------------------------
  //  単位変換
  //------------------------------------------------------------------------------------
  TimedDoubleSeq q_unit;
  q_unit.data.length(23);
  for(unsigned int i = 0; i < 2; i++) {
    q_unit.data[i] = q.data[i] / 1000.0;   // mm -> m   
  }
  for(unsigned int i = 2; i < 23; i++) {
    q_unit.data[i] = DEG2RAD(q.data[i]);   // deg -> rad
  }

  // engineではWAIST部分がZYXの順番なので逆転
  // (rx, ry, rz) -> (rz, ry, rx)
  double rx_in, ry_in, rz_in;
  rx_in = q_unit.data[3];  ry_in = q_unit.data[4];  rz_in = q_unit.data[5];
  q_unit.data[3] = rz_in;  q_unit.data[4] = ry_in;  q_unit.data[5] = rx_in;
  


  //------------------------------------------------------------------------------------
  //  update current model
  //------------------------------------------------------------------------------------
  // ここで現在の関節角度を計算
  if(engine.setCurAngle(q_unit) == false) {       // 現在の関節各度をセット
    cout << "error : [Redun] : setCurAngle" << endl;
  }
  if(firstCircle) {
    selectMode();          // 目標値のデフォルト値を現在値に
    firstCircle = false;
  }


  //------------------------------------------------------------------------------------
  //  get cur position
  //------------------------------------------------------------------------------------
  // m_xyzR_outに反映
  TimedDoubleSeq p = engine.getCurPosition();
  // m_redun_cur_outに反映
  TimedDoubleSeq redun = engine.getCurRedundant();


  //------------------------------------------------------------------------------------
  //  単位変換
  //------------------------------------------------------------------------------------
  // m_xyzR_outに反映
  for(unsigned int i = 0; i < 3; i++) {
    m_xyzR_out.data[i]       = p.data[i] * 1000.0;       // 右腕（位置） m -> mm
    m_xyzR_out.data[i+12]    = p.data[i+12] * 1000.0;    // 左腕（位置） m -> mm
  }
  for(unsigned int i = 3; i < 12; i++) {
    m_xyzR_out.data[i]       = p.data[i];                // 右腕（姿勢）
    m_xyzR_out.data[i+12]    = p.data[i+12];             // 左腕（姿勢）
  }
   
  // m_redun_cur_outに反映
  for(int i = 0; i < 2; i++) {
    m_redun_cur_out.data[i] = redun.data[i] * 1000.0;  // m -> mm
  }
  for(int i = 2; i < 8; i++) {
    m_redun_cur_out.data[i] = RAD2DEG(redun.data[i]);  // rad -> deg
  }

  // engineではWAIST部分がZYXの順番なので逆転
  // (rz, ry, rx) -> (rx, ry, rz)
  double rx_out, ry_out, rz_out;
  rx_out = m_redun_cur_out.data[5];  ry_out = m_redun_cur_out.data[4];  rz_out = m_redun_cur_out.data[3];
  m_redun_cur_out.data[3] = rx_out;  m_redun_cur_out.data[4] = ry_out;  m_redun_cur_out.data[5] = rz_out;

  //test
  //cout << "\t\tr_elbow : cur = " << m_redun_cur_out.data[6] << "   ref = " << RAD2DEG(m_redun_ref_in_keeper.data[6]);
  //cout << "\t\tl_elbow : cur = " << m_redun_cur_out.data[7] << "   ref = " << RAD2DEG(m_redun_ref_in_keeper.data[7]) << endl;
  
  //------------------------------------------------------------------------------------
  //  出力  
  //------------------------------------------------------------------------------------
  m_xyzR_outOut.write();
  m_redun_cur_outOut.write();



#ifdef LOGGING
  //------------------------------------------------------------------------------------
  //  LOG
  //------------------------------------------------------------------------------------
  // log_q_cur_in
  ofs3.precision(3);
  ofs3 << q.data[0];
  for(int i = 1; i < 23; i++){
    ofs3 << ", " << q.data[i];
  }
  ofs3 << std::endl;

  // log_xyzR_out
  ofs4.precision(3);
  ofs4 << m_xyzR_out.data[0];
  for(int i = 1; i < 24; i++){
    ofs4 << ", " << m_xyzR_out.data[i];
  }
  ofs4 << std::endl;

  // log_redun_cur_out
  ofs6.precision(3);
  ofs6 << m_redun_cur_out.data[0];
  for(int i = 1; i < 8; i++) {
    ofs6 << ", " << m_redun_cur_out.data[i];
  }
  ofs6 << std::endl;
#endif //LOGGING
}




void RedundancySolutionCore::_update_m_xyzR_in(const ConnectorInfo& info, const TimedDoubleSeq& d)
{
  if(d.data.length() != 24) {
    cout << "error : [Redun] : update m_xyzR_in" << endl;
    return;
  }
  

  //------------------------------------------------------------------------------------
  //  redun_ref_inのアップデート
  //------------------------------------------------------------------------------------
  //m_redun_ref_inIn.update();
  m_redun_ref_inIn.read();

  if(m_redun_ref_in.data.length() != 8) {
    cout << "error : [Redun] : update m_redun_ref_in" << endl;
    return;
  }


  //------------------------------------------------------------------------------------
  //  目標値の設定
  //------------------------------------------------------------------------------------
  setRefPosition(d);
  setRefRedundant(m_redun_ref_in);

 
  //------------------------------------------------------------------------------------
  //  目標関節角度の計算
  //------------------------------------------------------------------------------------
  if( engine.update(m_toRedundancy.m_mode, m_toRedundancy.m_type, m_toRedundancy.m_isHoldOtherArm) )
    m_status_out.data[0] = 1;
  else
    m_status_out.data[0] = 0;


  // for status port
  if( engine.condition_number > 0 )
    m_status_out.data[1] = engine.condition_number;
  else
    m_status_out.data[1] = 0.0;



  //------------------------------------------------------------------------------------
  //  get result angle
  //------------------------------------------------------------------------------------
  // 関節角の更新
  TimedDoubleSeq q = engine.getResultAngle();      // 結果の関節角度を取得
  for(unsigned int i = 0; i < q.data.length(); ++i) {
    if(isnan(q.data[i])) {
      cout << "error : [Redun] : q[" << i << "] = nan" << endl;
      //return;
    }
  } 
  

  //------------------------------------------------------------------------------------
  //  単位変換
  //------------------------------------------------------------------------------------
  // m_q_ref_outへ反映
  for(unsigned int i = 0; i < m_q_ref_out.data.length(); i++) {
    m_q_ref_out.data[i] = RAD2DEG(q.data[i]);                 // radian -> degree
  }
  m_q_ref_out.data[0] = q.data[0] * 1000;   // base x : m -> mm
  m_q_ref_out.data[1] = q.data[1] * 1000;   // base y : m -> mm

  // engineではWAIST部分がZYXの順番なので逆転
  // (rz, ry, rx) -> (rx, ry, rz)
  double rx_out, ry_out, rz_out;
  rx_out = m_q_ref_out.data[5];  ry_out = m_q_ref_out.data[4];  rz_out = m_q_ref_out.data[3];
  m_q_ref_out.data[5] = rz_out;  m_q_ref_out.data[4] = ry_out;  m_q_ref_out.data[3] = rx_out;

  
  //------------------------------------------------------------------------------------
  //  出力
  //------------------------------------------------------------------------------------
  m_status_outOut.write();
  m_q_ref_outOut.write();
  


#ifdef LOGGING
  //------------------------------------------------------------------------------------
  //  LOG
  //------------------------------------------------------------------------------------
  // log_xyzR_in
  ofs1.precision(3);
  ofs1 << d.data[0];
  for(int i=1; i<24; i++){
    ofs1 << ", " << d.data[i];
  }
  ofs1 << std::endl;

  // log_q_ref_out
  ofs2.precision(3);
  ofs2 << m_q_ref_out.data[0];
  for(int i=1; i<23; i++){
    ofs2 << ", " << m_q_ref_out.data[i];
  }
  ofs2 << std::endl;

  // log_redun_ref_in
  ofs5.precision(3);
  ofs5 << m_redun_ref_in.data[0];
  for(int i = 1; i < 8; i++) {
    ofs5 << ", " << m_redun_ref_in.data[i];
  }
  ofs5 << std::endl;
#endif //LOGGING
}



////////////////////////////////////////////////////////////////////////////////////////

extern "C"
{
 
  void RedundancySolutionCoreInit(RTC::Manager* manager)
  {
    coil::Properties profile(redundancysolutioncore_spec);
    manager->registerFactory(profile,
                             RTC::Create<RedundancySolutionCore>,
                             RTC::Delete<RedundancySolutionCore>);
  }
  
};


